@@ -1,15 +1,18 @@
 use crate::{
-    agent::{NonceFactory, ReplicaV2Transport},
+    agent::{NonceFactory, NonceGenerator, ReplicaV2Transport},
     identity::{anonymous::AnonymousIdentity, Identity},
 };
-use std::sync::Arc;
+use std::{sync::Arc, time::Duration};
 
 /// A configuration for an agent.
-pub struct AgentConfig {
-    pub nonce_factory: NonceFactory,
-    pub identity: Arc<dyn Identity + Send + Sync>,
-    pub ingress_expiry_duration: Option<std::time::Duration>,
-    pub transport: Option<Arc<dyn ReplicaV2Transport + Send + Sync>>,
+
+pub type AgentConfig = AgentConfigImpl<NonceFactory>;
+
+pub struct AgentConfigImpl<N: NonceGenerator> {
+    pub nonce_factory: N,
+    pub identity: Arc<dyn Identity>,
+    pub ingress_expiry_duration: Option<Duration>,
+    pub transport: Option<Arc<dyn ReplicaV2Transport>>,
 }
 
 impl Default for AgentConfig {
